import os
from typing import Dict
import logging
import torch
import tiinfer
import base64
import mosec

from diffusers import StableDiffusionPipeline

### 如果使用加速，需要导入如下代码
import tiacc_inference
TIACC_PATH = "model/tiacc"
### 如果使用加速，需要导入如上代码

# 未使用加速时模型路径
#MODEL_PATH = "."

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(process)d - %(levelname)s - %(filename)s:%(lineno)s - %(message)s"
)
sh = logging.StreamHandler()
sh.setFormatter(formatter)
logger.addHandler(sh)


class Infer(mosec.Worker):
    def __init__(self) -> None:
        super().__init__()
        # 获取当前所在目录，需要据此加载模型
        self.root_path = tiinfer.get_ti_model_dir()

        ### 原始模型加载,若使用加速模型，需要注释掉这段
        #model_id = os.path.join(self.root_path, MODEL_PATH)
        #pipe = StableDiffusionPipeline.from_pretrained(model_id, torch_dtype=torch.float16)
        #self.model = pipe.to("cuda")
        ### 原始模型加载,若使用加速模型，需要注释掉这段

        ### 加速模型,需要导入这段
        model_id = os.path.join(self.root_path, TIACC_PATH)
        pipe = tiacc_inference.load(model_id)[0]
        self.model = pipe
        ## 预热一下
        ### 加速模型,需要导入这段
        
        # 预热一下模型
        for i in range(2):
            img = self.model("a photo of an astronaut riding a horse on mars").images[0]
        torch.cuda.synchronize()

        try:
            self.uploader = CosUploader()
        except Exception as e:
            self.uploader = None
            logging.info(f'no cos uploader found: {e}')

    def forward(self, data: Dict) -> Dict:
        images = self.model(**data["draw_parameters"]).images
        result = []
        for img in images:
            url = None
            if self.uploader is not None:
                url = self.uploader.upload_image(img)
            if url is None or len(url) == 0:
                bs = bytes_from_image(img)
                url = base64.b64encode(bs).decode("ascii")
            result.append(url)
        return {"images": result}

tiinfer.append_worker(Infer, 1)

def bytes_from_image(image):
    import io
    buffer = io.BytesIO()
    image.save(buffer, format='PNG')
    return buffer.getvalue()


class CosUploader:
    def __init__(self):
        if os.getenv('COS_UPLOAD_ENABLE') is None:
            raise ValueError("environment COS_UPLOAD_ENABLE is not set")
        if os.getenv('COS_SECRET_ID') is None:
            raise ValueError("environment COS_SECRET_ID is not set")
        if os.getenv('COS_SECRET_KEY') is None:
            raise ValueError("environment COS_SECRET_KEY is not set")
        if os.getenv('COS_BUCKET') is None:
            raise ValueError("environment COS_BUCKET is not set")
        if os.getenv('COS_REGION') is None:
            raise ValueError("environment COS_REGION is not set")

        # 1. 设置用户属性, 包括 secret_id, secret_key, region 等。Appid 已在 CosConfig 中移除，请在参数 Bucket 中带上 Appid。Bucket 由 BucketName-Appid 组成
        # 用户的 SecretId，建议使用子账号密钥，授权遵循最小权限指引，降低使用风险。子账号密钥获取可参见 https://cloud.tencent.com/document/product/598/37140
        secret_id = os.getenv("COS_SECRET_ID")
        # 用户的 SecretKey，建议使用子账号密钥，授权遵循最小权限指引，降低使用风险。子账号密钥获取可参见 https://cloud.tencent.com/document/product/598/37140
        secret_key = os.getenv("COS_SECRET_KEY")
        # 替换为用户的 region，已创建桶归属的 region 可以在控制台查看，https://console.cloud.tencent.com/cos5/bucket
        # COS 支持的所有 region 列表参见 https://cloud.tencent.com/document/product/436/6224
        region = os.getenv("COS_REGION")
        # 如果使用永久密钥不需要填入 token，如果使用临时密钥需要填入，临时密钥生成和使用指引参见 https://cloud.tencent.com/document/product/436/14048
        token = None               
        # 指定使用 http/https 协议来访问 COS，默认为 https，可不填
        scheme = 'https'
        self.bucket = os.getenv('COS_BUCKET')
        self.prefix = os.getenv('COS_FILE_PREFIX', '')

        from qcloud_cos import CosConfig
        from qcloud_cos import CosS3Client
        config = CosConfig(Region=region, SecretId=secret_id, SecretKey=secret_key, Token=token, Scheme=scheme)
        self.client = CosS3Client(config)
    
    def upload_image(self, image, key: str=None) -> str:
        if key is None:
            import uuid
            key = str(uuid.uuid4()) + '.png'
        if len(self.prefix) > 0:
            key = os.path.join(self.prefix, key)
        from qcloud_cos.cos_exception import CosClientError, CosServiceError

        # 使用高级接口断点续传，失败重试时不会上传已成功的分块(这里重试10次)
        for i in range(0, 10):
            try:
                response = self.client.put_object(
                    Bucket=self.bucket,
                    Body=bytes_from_image(image),
                    Key=key)
                url = self.client.get_object_url(
                    Bucket=self.bucket,
                    Key=key,
                )
                return url
            except (CosClientError, CosServiceError) as e:
                logging.error(f"failed to upload image: {e}")
        return ""